pub mod shell {
    use std::process::Command;
    use std::io::stdin;
    use std::io::stdout;
    use std::io::Write;
    use std::io::ErrorKind;
    use std::path::Path;
    use std::env;
    use std::str;
    use std::process::Child;
    use std::process::Stdio;
    use std::fs::File;
    use std::fs::OpenOptions;
    use std::fs;
    use std::io::{self, prelude::*, BufReader};
    use users::{get_user_by_uid, get_current_uid};
    use gethostname::gethostname;
    use termion::event::Key;
    use termion::input::TermRead;
    use termion::raw::IntoRawMode;


    //used for home directory
    extern crate dirs;



    fn process_line(input: &str) -> String {
        let mut commands = input.trim().split(" | ").peekable();
        let mut previous_command = None;
        let mut resultat = String::new();

        while let Some(command) = commands.next()  {
            let mut parts = command.trim().split_whitespace();
            let mut command = parts.next().unwrap();
            let args = parts;

            match command {
                "cd" => {
                    let new_dir = args.peekable().peek().map_or("/", |x| *x);
                    let root = Path::new(new_dir);
                    if let Err(e) = env::set_current_dir(&root) {
                        eprintln!(" Err: {}", e);
                    }
                    previous_command = None;
                },
                "history" => {
                    let stdout = Stdio::inherit();
                    let res = get_history();
                    match res {
                        Ok(r) => return r,
                        Err(e) => eprintln!(" Err: {}", e),
                    }
                },
                command => {
                    if commands.peek().is_some() {
                        let stdin  = match previous_command {
                            None => Stdio::inherit(),
                            Some(o) => o,
                        };
                        let stdout = Stdio::piped();
                    
                        let output = Command::new(command)
                            .args(args)
                            .stdout(stdout)
                            .stdin(stdin)
                            .spawn();

                        match output {
                            Ok(o) => previous_command = Some(Stdio::from(o.stdout.unwrap())),
                            Err(e) => {
                                previous_command = None;
                                eprintln!(" Err: {}", e);
                            },
                        };
                    } else {
                        let stdin = match previous_command {
                            None => Stdio::inherit(),
                            Some(o) => o,
                        };
                        let output = Command::new(command)
                            .args(args)
                            .stdin(stdin)
                            .output()
                            .expect("???");
                        &mut resultat.push_str(str::from_utf8(&output.stdout)
                                                    .expect("Could not convert command to str")
                                                    );
                        previous_command = None;
                    }
                },
            };
        }
        return resultat;
    }


    fn write_to_history(command: &str) -> Result<(), std::io::Error> {
        //Write a command to history
        let filepath = dirs::home_dir().unwrap().join("history.rshell");
        let mut file = OpenOptions::new()
            .write(true)
            .append(true)
            .create(true)
            .open(&filepath)
            .unwrap();

        let number = match get_history_number() {
            Ok(n) => n + 1,
            Err(_) => 1
        };

        writeln!(file, "{} {}", number, command.trim())?;
        return Ok(())
    }

    fn get_history_number() -> Result<i32, std::io::Error> {
        //Get the latest number found in history file
        let filepath = dirs::home_dir().unwrap().join("history.rshell");
        let file = File::open(filepath)?;
        let mut reader = BufReader::new(file).lines();
        let myline = String::from(reader.last().unwrap_or(Ok(String::from("1")))?);
        let mut number = myline.split(' ').next().expect("???").parse().unwrap();
        return Ok(number);
    }

    fn get_history() -> Result<String, std::io::Error> {
        let filepath = dirs::home_dir().unwrap().join("history.rshell");
        let file = File::open(filepath)?;
        let contents = BufReader::new(file).lines();
        let mut res = String::new();
        for line in contents {
            let myline = format!("\r\n{}", line.unwrap());
            res.push_str(&myline);
        }
        return Ok(res);
    }

    fn get_hist_from_number(number: &i32) -> Option<String> {
        //Returns a command from its number in hist file
        let filepath = dirs::home_dir().unwrap().join("history.rshell");

        if filepath.exists() == false {
            return None;
        }

        let file = File::open(filepath)
            .expect("Error opening history file...");

        if file.metadata().unwrap().len() < 1 {
            return None;
        }

        let mut reader = BufReader::new(file).lines();

        for line in reader {
            let current_line = line
                .expect("Empty history file ? Please rm it");
            let mut s = current_line.split_whitespace();
            let n: i32 = s.next()
                .expect("Error reading a line in hist file")
                .parse()
                .unwrap();
            let c = s.next().expect("No command!");
            if &n == number {
                return Some(String::from(c));
            }
        }
        return None;
    }

    fn treat_history_callback(line: &str) -> Option<String> {
        let mut mystring = String::from(line);
        mystring = mystring.trim().to_string();
        let mut chars = mystring.chars();
        //Skip the !
        chars.next();
        let mynbr: i32 = chars.as_str().parse().unwrap();
        get_hist_from_number(&mynbr)
    }

    fn build_prompt() -> String {
        let user = String::from(
            get_user_by_uid(get_current_uid())
            .unwrap()
            .name()
            .to_str()
            .unwrap()
            );

        let host = String::from(
            gethostname::gethostname()
            .to_str()
            .unwrap()
            );

        let current_dir = String::from(
            env::current_dir()
            .unwrap()
            .to_str()
            .unwrap()
            );

        let s = current_dir.split('/');
        let dir_last = String::from(
            s.last()
            .unwrap()
            );

        let prompt = String::from(format!("[{}@{} in {}] ", user, host, dir_last));
        return prompt;
    }

    fn replace_signs(line: &String) -> String {
        let mut ayo = String::from(line);
        if ayo.contains('~') {
            let homedir = dirs::home_dir().unwrap().into_os_string().into_string().unwrap();
            let result = str::replace(&ayo, "~", &homedir);
            ayo = String::from(result);
        }
        
        if ayo.contains("!!") {
            let prev_comm = get_hist_from_number(&get_history_number().unwrap()).unwrap();
            let result = str::replace(&ayo, "!!", &prev_comm);
            ayo = String::from(result);
        }

        return ayo;
    }

    fn handle_input(line: &str) -> String {
        //history callback
        if (line.len() > 0) && (line.starts_with('!'))  {
            let command_found = treat_history_callback(line);
            match command_found {
                Some(c) => {
                    write_to_history(&c);
                    let outp = process_line(&c);
                    return outp;
                    },
                None => return String::from(" "),
            };
        //Regular command
        } else if line.len() > 0 {
            write_to_history(line);
            return process_line(line);
        //Nothing
        } else {
            //Command was empty, new line
            return String::from(" ");
        }
    }

    fn fake_tab(input: &str) -> String {
        let faketab = format!(" --This is a fake output for autocomplete from {input}-- ");
        //In reality, we would keep the input and append to it
        let aaa = String::from(faketab);
        return aaa;
    }

    pub fn run_raw() {
        let stdin = stdin();
        let mut stdout = stdout().into_raw_mode().unwrap();
        let prompt = build_prompt();
        let mut mycommand = String::new();

        //Initialize
        write!(stdout, "\r\n SquiShell (sqish)--- \r\n{}", prompt);
        stdout.flush();

        for c in stdin.keys() {

            match c.unwrap() {
                Key::Char('\t') => {
                    let res = fake_tab(&mycommand);
                    mycommand.clear();
                    mycommand.push_str(&res);
                    write!(stdout, "\r\n{}{}", prompt, res);
                    stdout.flush();
                }

                Key::Char('\n') => {
                    if (mycommand != String::from("\n")) && (mycommand != String::from("exit")) {
                        let comm = replace_signs(&mycommand);
                        let res = handle_input(&comm);
                        mycommand.clear();
                        //Proper printing
                        for line in res.split('\n') {
                            write!(stdout, "\r\n{}", line);
                        }
                        write!(stdout, "\r\n{}", prompt).unwrap();
                        stdout.flush();
                    } else if mycommand == String::from("exit") {
                        break;
                    } else {
                        write!(stdout, "\r\n{}", prompt).unwrap();
                        stdout.flush();
                    }
                },

                Key::Ctrl('q') => break,

                Key::Char(C) => {
                    mycommand.push(C);
                    write!(stdout, "{}", C).unwrap();
                },

                Key::Backspace => {
                    match mycommand.pop() {
                        Some(_) => {},
                        None => continue,
                    }
                    //Move cursor left
                    write!(stdout, "\x1b[D").unwrap();
                    //Delete one character there
                    write!(stdout, "\x1b[K").unwrap();
                    stdout.flush();
                },

                _ => (),
            }

            stdout.flush().unwrap();
        }

    }

//    pub fn run(){
//        loop {
//            print!("{}", build_prompt());
//            stdout().flush();
//            let mut input = String::new();
//            let bytes = stdin().read_line(&mut input).unwrap();
//            input = replace_signs(input);
//
//            //history callback
//            if (bytes > 0) && (input != String::from("\n")) && (input.starts_with('!'))  {
//                let command_found = treat_history_callback(&input);
//                match command_found {
//                    Some(c) => {
//                        write_to_history(&c);
//                        process_line(&c);
//                        },
//                    None => ()
//                };
//            //Regular command
//            } else if (bytes > 0) && (input != String::from("\n")) {
//                write_to_history(&input);
//                if process_line(input) {
//                    return
//                }
//            //Nothing
//            } else {
//                //Command was empty, new line
//                continue;
//            }
//        }
//    }













    #[cfg(test)]
    mod tests {
        use super::*;

        fn inittests() {
            let filepath = dirs::home_dir().unwrap().join("history.rshell");

            let file = File::create(&filepath).expect("Could not create test history");
            writeln!(&file, "1 ls");
            writeln!(&file, "2 pwd");
        }

        #[test]
        fn test_gethistnumber() {
            inittests();
            let mynumber = get_history_number().unwrap();
            assert_eq!(mynumber, 2);
        }

        #[test]
        fn test_gethistfromnumber() {
            inittests();
            let mycmd = get_hist_from_number(&1).unwrap();
            assert_eq!(mycmd, String::from("ls"));
        }

        #[test]
        fn test_history_callback() {
            inittests();
            let expected = String::from("ls");
            assert_eq!(treat_history_callback("!1"), Some(expected));
        }

        #[test]
        fn test_replace_signs() {
            inittests();
            let homedir = dirs::home_dir().unwrap().into_os_string().into_string().unwrap();
            assert_eq!(replace_signs(String::from("ls ~ && echo !!")), String::from(format!("ls {} && echo pwd", homedir)));
        }

        #[test]
        fn test_writehistline() {
            inittests();
            let towrite = String::from("pwd");
            write_to_history(&towrite);

            //Now check directly against the file
            let filepath = dirs::home_dir().unwrap().join("history.rshell");
            let file = File::open(filepath).expect("Error");
            let mut reader = BufReader::new(file).lines();
            let myline = String::from(reader.last().unwrap().expect("Error"));
            let expected = String::from("3 pwd");

            assert_eq!(myline.trim(), expected.trim()); 
        }
    }
}


